import json
import os
from urllib.parse import urlencode

import pytest
from flask import url_for

from .flask_app import create_app

# to find out how to write tests using flask, see https://github.com/graphql-python/flask-graphql/blob/master/tests/test_graphqlview.py
# to find out how to upload files in the tests, see https://github.com/lmcgartland/graphene-file-upload/blob/3f660ff17dc18d5bdf1c4008dccc33f528056000/tests/test_flask.py

# helper functions
def response_utf8_json(resp):
    return json.loads(resp.data.decode())


def response_json(response):
    return json.loads(response.data.decode())


def url_string(app, **url_params):  # pylint: disable=redefined-outer-name
    with app.test_request_context():
        string = url_for("graphql")

    if url_params:
        string += "?" + urlencode(url_params)

    return string


@pytest.fixture
def client():
    test_client = create_app().test_client()
    yield test_client


@pytest.fixture
def app():
    # import app factory pattern
    test_app = create_app()

    # pushes an application context manually
    ctx = test_app.app_context()
    ctx.push()
    return test_app


@pytest.mark.parametrize(
    "filename",
    [
        ("testfile3.png"),
        # ("testfile2.png"),  # fails with error "a bytes-like object is required, not 'str'". Probably because it is
        # just a simple txt file
        # ("testfile3.png")  # fails with error message "'charmap' codec can't decode byte 0x8d in position 65:
        # character maps to <undefined>", this is probably because the file has data which cannot be represented in unicode.
    ],
)
def test_single_file(client, filename: str):  # pylint: disable=redefined-outer-name
    query = """
mutation testMutation($imageFile: Upload!) {
  myfile: UploadInvoice(file: $imageFile) {
    bankDocument {
      kid {
        value
      }
      invoiceAccountNumber {
        value
      }
    }
  }
}
"""
    path = os.path.join(os.path.dirname(os.path.abspath(__file__)), filename)
    with open(path, "rb") as t_file:
        response = client.post(
            "/graphql",
            data={
                "operations": json.dumps(
                    {
                        "query": query,
                        "variables": {
                            "imageFile": None,
                        },
                    }
                ),
                "t_file": t_file,
                "map": json.dumps(
                    {
                        "t_file": ["variables.imageFile"],
                    }
                ),
            },
        )
    parsed_response = response_utf8_json(response)
    assert not "errors" in parsed_response
    assert response.status_code == 200


def test_hello_world(client, app):  # pylint: disable=redefined-outer-name
    response = client.get(url_string(app, query="{hello}"))
    parsed_response = response_utf8_json(response)

    assert not "errors" in parsed_response
    assert response.status_code == 200
    assert parsed_response == {"data": {"hello": "hello"}}
