from enum import Enum
from typing import List, Optional, Tuple, Union

from pdf_invoice_reader_common.interface import BankVoucher, Invoice, RawBankvoucher, RawInvoice

from .resolver.text_account_number_finder import TextAccountNumberFinder
from .resolver.text_date_finder import TextDateFinder
from .resolver.text_kid_finder import TextKidFinder
from .resolver.text_number_field_finder import TextNumberFieldFinder
from .resolver.text_principal_amount_finder import TextPrincipalAmountFinder


def first_or_none(nullable_list: Union[List[str], str]) -> Union[List[str], Optional[str]]:
    if not isinstance(nullable_list, list):
        return nullable_list
    return next(iter(nullable_list), None)


class FieldType(Enum):
    KID = 1
    DATE = 2
    ACCOUNT_NUMBER = 3
    NUMBER = 4
    PRINCIPAL_AMOUNT = 5


def run_validator(field_type, raw_values):
    for raw_value in sorted(raw_values, key=lambda x: x.confidence)[::-1]:
        raw_value = raw_value.value
        if field_type == FieldType.KID:
            value = first_or_none(TextKidFinder().find_in_text(raw_value))
        elif field_type == FieldType.DATE:
            value = first_or_none(TextDateFinder().find_in_text(raw_value))
        elif field_type == FieldType.ACCOUNT_NUMBER:
            value = first_or_none(TextAccountNumberFinder().find_in_text(raw_value))
        elif field_type == FieldType.NUMBER:
            value = first_or_none(TextNumberFieldFinder().find_in_text(raw_value))
        elif field_type == FieldType.PRINCIPAL_AMOUNT:
            value = first_or_none(TextPrincipalAmountFinder().find_in_text(raw_value))
        else:
            value = first_or_none(raw_value)
        if value is not None:
            return value
    return None


def parse_raw_invoice(raw_invoice: RawInvoice) -> Tuple[RawInvoice, Invoice]:
    clean_invoice = Invoice()
    if len(raw_invoice.kid) > 0:
        raw_invoice.kid[0].parsed_value = run_validator(FieldType.KID, raw_invoice.kid)
    if len(raw_invoice.invoice_account_number) > 0:
        raw_invoice.invoice_account_number[0].parsed_value = run_validator(
            FieldType.ACCOUNT_NUMBER, raw_invoice.invoice_account_number
        )
    if len(raw_invoice.invoice_number) > 0:
        raw_invoice.invoice_number[0].parsed_value = run_validator(FieldType.NUMBER, raw_invoice.invoice_number)
    if len(raw_invoice.customer_number) > 0:
        raw_invoice.customer_number[0].parsed_value = run_validator(FieldType.NUMBER, raw_invoice.customer_number)
    if len(raw_invoice.date) > 0:
        raw_invoice.date[0].parsed_value = run_validator(FieldType.DATE, raw_invoice.date)
    if len(raw_invoice.due_date) > 0:
        raw_invoice.due_date[0].parsed_value = run_validator(FieldType.DATE, raw_invoice.due_date)
    if len(raw_invoice.principal_amount) > 0:
        raw_invoice.principal_amount[0].parsed_value = run_validator(
            FieldType.PRINCIPAL_AMOUNT, raw_invoice.principal_amount
        )

    for field in list(map(lambda x: x, raw_invoice.__dataclass_fields__)):
        values = getattr(raw_invoice, field)
        if values is None or len(values) == 0:
            continue
        values = values[0]
        values.value = values.parsed_value
        setattr(clean_invoice, field, values)

    return raw_invoice, clean_invoice


def extract_bankvoucher_fields_within_entry_box(raw_bankvoucher, entrybox):
    clean_bankvoucher = BankVoucher()
    for field in list(map(lambda x: x, raw_bankvoucher.__dataclass_fields__)):
        validator = None
        if field == "date":
            validator = TextDateFinder()
        elif field == "to_account_number":
            validator = TextAccountNumberFinder()
        elif field == "from_account_number":
            validator = TextAccountNumberFinder()
        elif field == "principal_amount":
            validator = TextPrincipalAmountFinder()

        for values in getattr(raw_bankvoucher, field):
            current_value = getattr(clean_bankvoucher, field)
            if entrybox is None and not isinstance(current_value, list) and current_value.value is not None:
                continue
            values.value = (
                first_or_none(validator.find_in_text(values.value))
                if validator is not None and values.value is not None
                else values.value
            )
            if entrybox is not None:
                value_box = values.bbox
                entry_box = entrybox.bbox
                within_window = entry_box[1] < value_box[1] and (
                    (value_box[3] < entry_box[3]) or abs(value_box[3] - entry_box[3]) < 25
                )
                if within_window and values.value is not None:
                    setattr(clean_bankvoucher, field, values)
            else:
                setattr(clean_bankvoucher, field, values)
    return clean_bankvoucher


def parse_raw_bankvoucher(raw_bankvoucher: RawBankvoucher) -> Tuple[RawBankvoucher, BankVoucher]:
    if len(raw_bankvoucher.entrybox) > 1:
        return raw_bankvoucher, [
            extract_bankvoucher_fields_within_entry_box(raw_bankvoucher, entrybox)
            for entrybox in raw_bankvoucher.entrybox
        ]
    return raw_bankvoucher, [extract_bankvoucher_fields_within_entry_box(raw_bankvoucher, None)]


def parse_document(
    document: Union[RawBankvoucher, RawInvoice]
) -> Union[Tuple[RawBankvoucher, List[BankVoucher]], Tuple[RawInvoice, List[Invoice]]]:
    assert isinstance(document, (RawBankvoucher, RawInvoice)), "wrong input type"
    if isinstance(document, RawBankvoucher):
        return parse_raw_bankvoucher(document)
    raw_invoice, clean_invoice = parse_raw_invoice(document)
    return raw_invoice, [
        clean_invoice,
    ]
