import glob
import io
import os
import random
import time
from copy import deepcopy

import cv2
import fitz
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
import torch
import torchvision.transforms as transforms
from matplotlib.ticker import NullLocator
from pdf_invoice_reader_common.interface import Field, FieldName, RawBankvoucher, RawInvoice
from PIL import Image
from torch.autograd import Variable

from .default import data_config
from .default import data_config as default_data_config
from .default import model_def
from .eval_model import extract_text_from_image
from .models import Darknet
from .preprocess import preprocess_input
from .utils.datasets import resize
from .utils.parse_config import parse_data_config
from .utils.utils import load_classes, non_max_suppression, rescale_boxes, weights_init_normal

parsed_data_config = parse_data_config(data_config)
current_dir = os.path.dirname(os.path.abspath(__file__))
root_path = os.path.join(current_dir, parsed_data_config["root_folder"])
class_names = os.path.join(root_path, "classes.names")

classes = load_classes(class_names)


def preprocess_image(image, ignored_level, with_kernel=True, with_median=False):
    image = np.array(image)
    image[image > ignored_level]
    image[image == ignored_level] = 255
    if with_kernel:
        kernel = np.ones((1, 1), np.uint8)
        image = cv2.morphologyEx(image, cv2.MORPH_CLOSE, kernel)
    if with_median:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        #        ret,thresh_img = cv2.threshold(image, 0, 255, cv2.THRESH_BINARY_INV|cv2.THRESH_OTSU)
        _, thresh_img = cv2.threshold(image, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)
        image = thresh_img
    return image


class FieldBox:
    def __init__(self, bbox, confidence):
        self.bbox = bbox
        self.confidence = confidence


class Detect:
    def __init__(
        self, data_config=None, images=None, prefix="", use_test_dir=False, no_mark_class_label=False, load_model=True
    ):
        if load_model:
            self.debug = False
            if data_config is None:
                data_config = default_data_config
            self.prefix = prefix
            parsed_data_config = parse_data_config(data_config)
            self.current_dir = os.path.dirname(os.path.abspath(__file__))
            root_path = os.path.join(self.current_dir, parsed_data_config["root_folder"])
            class_names = os.path.join(root_path, "classes.names")
            if use_test_dir:
                self.image_path = os.path.join(root_path, "test-images")
            elif images is None:
                self.image_path = os.path.join(root_path, "images")
            else:
                assert os.path.isdir(images), "images path should be directory"
                self.image_path = images
            self.classes = load_classes(class_names)
            self.no_mark_class_label = no_mark_class_label

            device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
            checkpoint_path = os.path.join(
                self.current_dir, "checkpoints/{prefix}_latest_checkpoint.pth".format(prefix=self.prefix)
            )
            self.model = Darknet(model_def).to(device)
            self.model.apply(weights_init_normal)
            self.model.load_state_dict(torch.load(checkpoint_path))
            modification_time = os.path.getmtime(checkpoint_path)
            modification_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(modification_time))
            self._debug("checkpoint_path last modified time : ", modification_time)

    def _debug(self, *args):
        if self.debug:
            print(args)

    def plot_document(self, path, detections):
        plt.clf()
        cmap = plt.get_cmap("tab20b")
        colors = [cmap(i) for i in np.linspace(0, 1, 20)]

        img = np.array(self._open_image(self._get_image_file(path)))

        plt.figure(figsize=(10, 10))
        fig, ax = plt.subplots(1, 2)
        fig.set_size_inches(100, 100)

        ax[0].imshow(img)
        ax[1].imshow(img)

        unique_labels = detections[:, -1].cpu().unique()
        n_cls_preds = len(unique_labels)
        bbox_colors = random.sample(colors, n_cls_preds)

        image_copy = deepcopy(img)
        for x1, y1, x2, y2, _, _, cls_pred in detections:
            box_w = int(x2 - x1)
            box_h = int(y2 - y1)
            x1 = int(x1)
            y1 = int(y1)

            if not self.no_mark_class_label:
                color = bbox_colors[int(np.where(unique_labels == int(cls_pred))[0])]
                bbox = patches.Rectangle((x1, y1), box_w, box_h, linewidth=2, edgecolor=color, facecolor="none")

                ax[1].add_patch(bbox)
                ax[1].text(
                    x1,
                    y1,
                    s=self.classes[int(cls_pred)],
                    color="white",
                    fontsize=12,
                    verticalalignment="top",
                    bbox={"color": color, "pad": 0},
                )
            else:
                image_copy[y1 : y1 + box_h, x1 : x1 + box_w, :] = 0
        if self.no_mark_class_label:
            ax[1].imshow(image_copy)
        for i in range(2):
            ax[i].axis("off")
            ax[i].xaxis.set_major_locator(NullLocator())
            ax[i].yaxis.set_major_locator(NullLocator())
        return plt, fig

    def clean_plot_document(self, path, bbox):
        plt.clf()
        cmap = plt.get_cmap("tab20b")
        colors = [cmap(i) for i in np.linspace(0, 1, 20)]

        img = np.array(self._open_image(self._get_image_file(path)))

        plt.figure(figsize=(10, 10))
        fig, ax = plt.subplots(1, 1)
        fig.set_size_inches(100, 100)

        ax.imshow(img)

        field_id = {}
        for x in bbox:
            field_id[x[-1]] = len(field_id)
        n_cls_preds = len(field_id) + 1
        bbox_colors = random.sample(colors, n_cls_preds)

        deepcopy(img)
        for _, (x1, y1, x2, y2, field_name) in enumerate(bbox):
            box_w = int(x2 - x1)
            box_h = int(y2 - y1)
            x1 = int(x1)
            y1 = int(y1)

            if not self.no_mark_class_label:
                color = bbox_colors[field_id[field_name]]
                bbox = patches.Rectangle((x1, y1), box_w, box_h, linewidth=2, edgecolor=color, facecolor="none")

                ax.add_patch(bbox)
                ax.text(
                    x1,
                    y1,
                    s=field_name,
                    color="white",
                    fontsize=12,
                    verticalalignment="top",
                    bbox={"color": color, "pad": 0},
                )
        ax.axis("off")
        ax.xaxis.set_major_locator(NullLocator())
        ax.yaxis.set_major_locator(NullLocator())
        return plt, fig

    def detect(self, input_imge, img_size, image, conf_thres=0.75, nms_thres=0.45, debug=False):
        Tensor = torch.cuda.FloatTensor if torch.cuda.is_available() else torch.FloatTensor

        input_imgs = Variable(input_imge.type(Tensor)).permute(2, 0, 1)
        input_imgs = resize(input_imgs, img_size)
        input_imgs = input_imgs.reshape((1,) + input_imgs.shape)

        fields_masks = {}
        field_bbox = {}
        mask = np.zeros(image.shape[:-1])
        raw_document = RawInvoice() if not "bankvoucher" in self.prefix else RawBankvoucher()
        raw_detection = None

        with torch.no_grad():
            detections = self.model(input_imgs)
            detections = non_max_suppression(detections, conf_thres, nms_thres)
            if detections[0] is None:
                raise Exception("something went wrong ?")

            raw_detection = detections[0].clone()
            self._debug(image.shape[:2])
            detections = rescale_boxes(raw_detection, img_size, image.shape[:2])
            for x1, y1, x2, y2, conf, cls_conf, cls_pred in detections:
                self._debug((int(cls_pred), x1, y1, x2, y2, conf, cls_conf, self.classes[int(cls_pred)]))
                field_key = self.classes[int(cls_pred)]

                if field_key not in fields_masks:
                    fields_masks[field_key] = [mask.copy()]
                    field_bbox[field_key] = []
                else:
                    fields_masks[field_key].append(mask.copy())

                x1 -= 10
                x2 += 10

                x1 = int(x1)
                x2 = int(x2)
                y1 = int(y1)
                y2 = int(y2)

                fields_masks[field_key][-1][y1:y2, x1:x2] = 255
                field_bbox[field_key].append(FieldBox([x1, y1, x2, y2], confidence=(cls_conf + conf) / 2))

        for (field_key, field_masks) in fields_masks.items():
            for index, field_mask in enumerate(field_masks):
                ignored_level = 8
                applied_mask = image.copy()
                field_mask[field_mask > 0] = 1
                applied_mask[~field_mask.astype(bool), :] = 0
                applied_mask[~field_mask.astype(bool), :] = ignored_level
                applied_mask = transforms.ToPILImage()(preprocess_image(applied_mask, ignored_level))

                bbox = field_bbox[field_key][index].bbox
                confidence = field_bbox[field_key][index].confidence

                applied_mask = applied_mask.crop(bbox)
                basewidth = 300
                wpercent = basewidth / float(applied_mask.size[0])
                hsize = int((float(applied_mask.size[1]) * float(wpercent)))
                applied_mask = applied_mask.resize((basewidth, hsize), Image.ANTIALIAS)

                teseract_text = extract_text_from_image(applied_mask)
                teseract_text_2 = extract_text_from_image(
                    transforms.ToPILImage()(
                        preprocess_image(applied_mask, ignored_level, with_kernel=False, with_median=True)
                    )
                )
                self._debug(FieldName(field_key), teseract_text, teseract_text_2)

                path = os.path.join(
                    os.path.dirname(os.path.abspath(__file__)),
                    "{fieldkey}_ {index}.png".format(fieldkey=field_key, index=index),
                )
                applied_mask.save(path)
                transforms.ToPILImage()(
                    preprocess_image(applied_mask, ignored_level, with_kernel=False, with_median=False)
                ).save(path.replace(".png", "_v2.png"), quality=100)
                self._debug(path)
                self._debug(path.replace(".png", "_v2.png"))

                if FieldName(field_key) == FieldName.kid:
                    raw_document.kid.append(Field(teseract_text, confidence=confidence, bbox=bbox))
                elif FieldName(field_key) == FieldName.due_date:
                    raw_document.due_date.append(Field(teseract_text, confidence=confidence, bbox=bbox))
                elif FieldName(field_key) == FieldName.date:
                    raw_document.date.append(Field(teseract_text, confidence=confidence, bbox=bbox))
                elif FieldName(field_key) == FieldName.invoice_account_number:
                    raw_document.invoice_account_number.append(Field(teseract_text, confidence=confidence, bbox=bbox))
                elif FieldName(field_key) == FieldName.invoice_number:
                    raw_document.invoice_number.append(Field(teseract_text, confidence=confidence, bbox=bbox))
                elif FieldName(field_key) == FieldName.customer_number:
                    raw_document.customer_number.append(Field(teseract_text, confidence=confidence, bbox=bbox))
                elif FieldName(field_key) == FieldName.principal_amount:
                    raw_document.principal_amount.append(Field(teseract_text, confidence=confidence, bbox=bbox))
                # bankvoucher data
                elif FieldName(field_key) == FieldName.payer:
                    raw_document.payer.append(Field(teseract_text, confidence=confidence, bbox=bbox))
                elif FieldName(field_key) == FieldName.creditor:
                    raw_document.creditor.append(Field(teseract_text, confidence=confidence, bbox=bbox))
                elif FieldName(field_key) == FieldName.from_account_number:
                    raw_document.from_account_number.append(Field(teseract_text, confidence=confidence, bbox=bbox))
                elif FieldName(field_key) == FieldName.to_account_number:
                    raw_document.to_account_number.append(Field(teseract_text, confidence=confidence, bbox=bbox))
                elif FieldName(field_key) == FieldName.message:
                    raw_document.message.append(Field(teseract_text, confidence=confidence, bbox=bbox))
                elif FieldName(field_key) == FieldName.entrybox:
                    raw_document.entrybox.append(Field(teseract_text, confidence=confidence, bbox=bbox))
        raw_detection = detections
        return raw_document, detections, raw_detection

    def get_raw_yolo_format(self, image_file, normalized=False):
        _, _, raw_detection, _ = self.get_yolo_format(image_file)
        yolo_format = []
        for x1, y1, x2, y2, _, _, cls_pred in raw_detection:
            if not normalized:
                yolo_format.append([int(cls_pred), float(x1), float(x2), float(y1), float(y2)])
            else:
                x, y, w, h = self._convert_box_to_yolo_format([1030, 1030], [x1, x2, y1, y2])
                yolo_format.append([int(cls_pred), float(x), float(y), float(w), float(h)])
        return yolo_format

    def get_yolo_format(self, image_file, debug=False):
        self.debug = debug
        if type(image_file) == str:
            image = np.array(self._open_image(image_file).convert("RGB"), dtype=np.uint8)
        else:
            image = np.array(image_file.convert("RGB"), dtype=np.uint8)
        input_imgs = torch.from_numpy(image)
        raw_document, detections, raw_detection = self.detect(
            input_imgs, img_size=1030, conf_thres=0.3, image=image, debug=debug
        )
        return raw_document, detections, raw_detection, input_imgs

    def get_document_from_path(self, image_file, debug=False, plot=True, prefix=""):
        self.debug = debug
        image_file = self._get_image_file(image_file)
        return self.get_document_from_image(image_file, debug, plot)

    def get_document_from_image(self, image_file, debug=False, plot=False, prefix=""):
        self.debug = debug
        raw_document, detections, _, _ = self.get_yolo_format(image_file, debug)
        if plot:
            path = os.path.join(os.path.dirname(__file__), "test_output", prefix + os.path.basename(image_file))
            self.create_document_plot(image_file, detections, path)
        return raw_document

    def create_document_plot(self, image_file, detections, output_path):
        plt, fig = self.plot_document(image_file, detections)
        plt.savefig(
            output_path,
            bbox_inches="tight",
            pad_inches=0.0,
            dpi=30,
        )
        plt.cla()
        plt.close("all")

    def get_plot(self, image_file):
        image_file = self._get_image_file(image_file)
        _, detections, _, _ = self.get_yolo_format(image_file)
        plt, _ = self.plot_document(image_file, detections)
        return plt

    def predict_random_image(self, debug, plot):
        files = self._get_images()
        return self.get_document_from_path(files[random.randint(0, len(files) - 1)], debug=debug, plot=plot)

    def predict_all_files(self):
        files = self._get_images()
        output = []
        for pdf_path in files:
            try:
                img_path = self._get_image_file(pdf_path)
                self._debug(img_path)
                output.append(self.get_document_from_path(img_path, plot=True))
            except Exception as e:
                self._debug(e)
        return output

    def predict_all_files_outside_data_config(self, outside_directory):
        output = []
        root_files = list(map(lambda x: os.path.basename(x), self._get_images()))
        outside_directory = self._get_images(outside_directory)
        outside_directory_filtered = [i for i in outside_directory if os.path.basename(i) not in root_files]

        for file in outside_directory_filtered:
            try:
                img_path = self._get_image_file(file)
                image_file = self._get_image_file(img_path)
                output_path = os.path.join(os.path.dirname(__file__), "label_output", os.path.basename(image_file))
                if os.path.isfile(output_path):
                    continue
                _, detections, _, _ = self.get_yolo_format(image_file, False)
                self.create_document_plot(
                    image_file,
                    detections,
                    output_path=output_path,
                )
                img = np.array(self._open_image(img_path))
                img = Image.fromarray(img)
                img.save(output_path.replace(".png", "_label_file.png"))

                txt_path = output_path.replace(".png", ".txt")
                with open(txt_path, "w") as file:
                    for x1, y1, x2, y2, _, _, cls_pred in detections:
                        class_id = int(cls_pred)
                        x, y, w, h = self._convert_box_to_yolo_format([1030, 1030], [x1, x2, y1, y2])
                        file.write("{class_id} {x} {y} {w} {h}\n".format(class_id=class_id, x=x, y=y, w=w, h=h))
            except Exception as e:
                print(e)
        return output

    def _get_image_file(self, pdf_or_image, is_pdf=False):
        page_index = 0
        if pdf_or_image.endswith(".pdf") or is_pdf:
            img_path = pdf_or_image.replace(".pdf", ".png")
            if not img_path.endswith(".png"):
                img_path += ".png"
            doc = fitz.open(pdf_or_image)
            page = doc[page_index]
            scale = 150 / 72
            page_to_image = self._open_image(
                io.BytesIO(page.get_pixmap(matrix=fitz.Matrix(scale, scale)).getImageData("png"))
            )
            page_to_image.save(img_path)
            return img_path
        return pdf_or_image

    def _open_image(self, input_path):
        return preprocess_input(Image.open(input_path))

    def _get_images(self, image_path=None):
        root_path = self.image_path if image_path is None else image_path
        return glob.glob(os.path.join(root_path, "*.pdf")) + glob.glob(os.path.join(root_path, "*.png"))

    def _convert_box_to_yolo_format(self, size, box):
        dw = 1.0 / size[0]
        dh = 1.0 / size[1]
        x = (box[0] + box[1]) / 2.0
        y = (box[2] + box[3]) / 2.0
        w = box[1] - box[0]
        h = box[3] - box[2]
        x = x * dw
        w = w * dw
        y = y * dh
        h = h * dh

        return (x, y, w, h)
