import glob
import json
import os
import random
import warnings

import numpy as np
from PIL import Image, ImageFile

ImageFile.LOAD_TRUNCATED_IMAGES = True

import random

import torch
import torch.nn.functional as F
from pdf_invoice_reader_common.filter import should_skip_file
from torch.utils.data import Dataset

from ..preprocess import preprocess_input
from ..weigthed_dataloader import WeightedDataloder

current = os.path.dirname(os.path.abspath(__file__))
root_path = os.path.join(current, "../../../pdf_invoice_reader/datasets")


def pad_to_square(img, pad_value):
    c, h, w = img.shape
    dim_diff = np.abs(h - w)
    # (upper / left) padding and (lower / right) padding
    pad1, pad2 = dim_diff // 2, dim_diff - dim_diff // 2
    # Determine padding
    pad = (0, 0, pad1, pad2) if h <= w else (pad1, pad2, 0, 0)
    # Add padding
    img = F.pad(img, pad, "constant", value=pad_value)

    return img, pad


def resize(image, size):
    image = F.interpolate(image.unsqueeze(0), size=size, mode="nearest").squeeze(0)
    return image


class ImageFolder(Dataset):
    def __init__(self, folder_path, file_name=None, transform=None, max_files=1):
        self.files = sorted(glob.glob("%s/*.png" % folder_path)) if file_name is None else [file_name]
        random.shuffle(self.files)
        self.files = self.files[:max_files]
        self.transform = transform

    def get_json_from_file(self, path):
        creditor = os.path.dirname(path).split("/")[-1]
        invoice_id = os.path.basename(path).replace("_normal", "").split(".")[0]
        path = os.path.join(root_path, "jsons")
        json_path = os.path.join(path, creditor, invoice_id + ".json")
        with open(json_path, "r") as file:
            return json.loads(file.read())

    def __getitem__(self, index):

        img_path = self.files[index % len(self.files)]
        img = np.array(Image.open(img_path).convert("RGB"), dtype=np.uint8)

        # Label Placeholder
        boxes = np.zeros((1, 5))

        # Apply transforms
        if self.transform:
            img, _ = self.transform((img, boxes))

        return img_path, img

    def __len__(self):
        return len(self.files)


def get_train_and_validate_dataset(image_path, split, opt, transform, max_files=None, filters=None):
    images = get_trainable_files(image_path, opt, filters)
    dataset_size = len(images) if max_files is None else min(len(images), max_files)

    first = int(split * dataset_size)
    skip = first
    left = dataset_size - first
    train = ListDataset(images[:first], multiscale=opt.multiscale_training, img_size=opt.img_size, transform=transform)
    if left == 0:
        return train, train

    test = ListDataset(
        images[skip:][:first],
        multiscale=opt.multiscale_training,
        img_size=opt.img_size,
        transform=transform,
    )
    return train, test


get_label = lambda x: x.replace("images", "labels").replace(".png", ".txt").replace(".jpg", ".txt")


def get_trainable_files(image_path, opt, filters):
    img_files = glob.glob(os.path.join(image_path, "*.png"))
    img_files = list(
        filter(
            lambda x: (
                (
                    ("augmentation") in x
                    or (not "_" in os.path.basename(x) and os.path.isfile(get_label(x)))
                    or (not "_" in os.path.basename(x) and os.path.isfile(get_label(x)))
                )
                or (
                    os.path.isfile(get_label(x))
                    and "_" in os.path.basename(x)
                    and not should_skip_file(x)
                    and os.path.basename(x).split(".")[0].replace("_", "").isnumeric()
                    and "debtor" in get_json_from_path(x)
                )
            ),
            img_files,
        )
    )
    if filters is not None:
        img_files = list(filter(lambda x: os.path.basename(x) in filters, img_files))
        img_files += img_files

    if opt is not None and opt.weighted:
        group_creditors = {}
        group_creditors_json_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "dataloader_weigths.json")
        if os.path.isfile(group_creditors_json_path):
            with open(group_creditors_json_path, "r") as file:
                group_creditors = json.loads(file.read())
            img_files = WeightedDataloder(group_creditors, img_files).get_new_batch()

    return img_files


def get_json_from_path(path):
    file_name = os.path.basename(path)
    creditor = file_name.split("_")[0]
    invoice_id = file_name.split("_")[1].split(".")[0]
    path = os.path.join(root_path, "jsons")
    json_path = os.path.join(path, creditor, invoice_id + ".json")
    if not os.path.isfile(json_path):
        return {}
    with open(json_path, "r") as file:
        return json.loads(file.read())


class ListDataset(Dataset):
    def __init__(self, img_files, img_size=416, multiscale=True, transform=None, image_only=False):
        self.img_files = img_files
        self.label_files = [get_label(path) for path in self.img_files]

        self.img_size = img_size
        self.max_objects = 100
        self.multiscale = multiscale
        self.min_size = self.img_size - 3 * 32
        self.max_size = self.img_size + 3 * 32
        self.batch_count = 0
        self.transform = transform
        self.image_only = image_only

    def get_creditor_ids(self):
        files = list(set(list(map(lambda x: os.path.basename(x).split("_")[0], self.img_files))))
        return files

    def get_json_from_file(self, path):
        return get_json_from_path(path)

    def __getitem__(self, index):

        # ---------
        #  Image
        # ---------
        try:

            img_path = self.img_files[index % len(self.img_files)].rstrip()
            input_image = preprocess_input(Image.open(img_path).convert("RGB"))
            assert abs(input_image.width - input_image.height) < 5, "no ! The image should be square for best results."
            img = np.array(input_image, dtype=np.uint8)
        except Exception as e:
            print(f"Could not read image '{img_path}'.")
            print(e)
            return

        if self.image_only:
            img = self.transform((img, None))
            return img

        # ---------
        #  Label
        # ---------
        try:
            label_path = self.label_files[index % len(self.img_files)].rstrip()

            # Ignore warning if file is empty
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                boxes = np.loadtxt(label_path).reshape(-1, 5)
        except Exception:
            print(f"Could not read label '{label_path}'.")
            return

        # -----------
        #  Transform
        # -----------
        bb_targets = None
        if self.transform:
            try:
                img, bb_targets = self.transform((img, boxes))
            except:
                print(f"Could not apply transform.")
                return

        return img_path, img, bb_targets

    def collate_fn(self, batch):
        self.batch_count += 1

        # Drop invalid images
        batch = [data for data in batch if data is not None]

        paths, imgs, bb_targets = list(zip(*batch))

        # Selects new image size every tenth batch
        if self.multiscale and self.batch_count % 10 == 0:
            self.img_size = random.choice(range(self.min_size, self.max_size + 1, 32))

        # Resize images to input shape
        imgs = torch.stack([resize(img, self.img_size) for img in imgs])

        # Add sample index to targets
        for i, boxes in enumerate(bb_targets):
            boxes[:, 0] = i
        bb_targets = torch.cat(bb_targets, 0)

        return paths, imgs, bb_targets

    def __len__(self):
        return len(self.img_files)
